#include <interrupt.h>
#include <io.h>
#include <eeprom.h>
#include <string.h>

#include "my_defs.h"
#include "memory.h"
#include "serial.h"
#include "tools.h"

#include "ide.h"
#include "lcd.h"				 
#include "fat.h"
#include "mas.h"

//number of entries in dir buffer
unsigned char DirCount;
unsigned int  SaveOffset;

unsigned long DataStart;

//unsigned char ClusterSize;
unsigned long FAT_START;
unsigned long DIR_START;		
unsigned int Fat32Length;		//Anzahl in Sektoren
unsigned long first_cluster_in_cache;
unsigned long last_cluster;

unsigned long PlaylistStartCluster;
unsigned long PlaylistLaenge;

extern volatile unsigned char ZeitFenster;
extern volatile unsigned char Time;
extern void PumpData(void);
	
void get_fat(unsigned long cluster);
void read_lba(unsigned long ,unsigned char, unsigned char*);
unsigned char inline read_fat_sector(unsigned long);
void dir( unsigned long);
unsigned long cluster_to_sector( unsigned long);
unsigned long get_next_cluster( unsigned long);
void update_buf(unsigned char*);

/***********************************************************************************
- IDE CMD ausfueren; IMMER LESEN ! Nur LBA!
- Daten werden ausgelesen und nach "in_buffer" geschrieben
***********************************************************************************/
void read_lba( unsigned long lba, unsigned char sec_count, unsigned char *in_buffer)
{
unsigned int dummy=0;

	while( (*CB_STAT & CB_STAT_BSY) == CB_STAT_BSY);
	*CB_DH = ((lba>>24) & 0x0f) | DEVICE | 0x40;

	//wait for DRDY == 1 && SEEK_COMPLETE==1
	while((*CB_STAT & ( CB_STAT_BSY | CB_STAT_RDY | CB_STAT_SKC ))	!= (CB_STAT_RDY | CB_STAT_SKC));
	*CB_DC = CB_DC_HD15 | CB_DC_NIEN ;
	*CB_FR = 0;
	*CB_SC = sec_count;
	*CB_SN = lba;	
	*CB_CL = lba>>8;
	*CB_CH = lba>>16;
	*CB_CMD = 0x20;
		
	while(sec_count)
	{
		while (((*CB_STAT & (CB_STAT_DRQ | CB_STAT_BSY )) == CB_STAT_DRQ) && (sec_count))
		{
			in_buffer[dummy++] = *CB_DATA;
			in_buffer[dummy++] = *DATA_H;
			if ((dummy & 0x1ff) ==0) sec_count--;	
		}
	}
}

/***********************************************************************************
- liest die erste Partition auf der Platte 
- initialisiert globale Variablen fr den FAT32-Zugriff
***********************************************************************************/
void inline init_fat( void)
{
unsigned long first_sec;

	read_lba(0, 1, (unsigned char*) SECTORBUFFER_ADR);			//read bootsector
	first_sec = *(unsigned long*) (SECTORBUFFER_ADR + PAR_TABLE_OFFSET +8);
	read_lba(first_sec, 1, (unsigned char*) SECTORBUFFER_ADR);	//read bootblock

//	ClusterSize	= SectorBuffer[13];
#define 	ClusterSize   16

	//Anfang berechnen
	FAT_START	     = first_sec + *(unsigned int*) (SECTORBUFFER_ADR + 14);
	Fat32Length      = *(unsigned int*) (SECTORBUFFER_ADR +36);
	DIR_START		 = FAT_START + Fat32Length * 2;
	DataStart 		 = DIR_START + *(unsigned char*)  (SECTORBUFFER_ADR +17);	
}

/***********************************************************************************
- liest einen Sektor von einem Verzeichnis ein
- schreibt in den DirBuffer jede "normale" Datei und jedes "normale" Unterverzeichnis
- im DirBuffer landet 
	- unsigned long  StartCluster
	- unsigned long  Lnge
	- String mit dem Namen (Null-terminiert)
	- bei Verzeichnissen ein DirMark als erstes Zeichen im Namen
- return 0, wenn das Ende des Verzeichnisses nicht erreicht wurde, sonst return 1
***********************************************************************************/
unsigned char inline read_fat_sector( unsigned long d_sector)
{

	unsigned char dummy;
//	struct msdos_dir_entry *entry;
//	struct msdos_dir_slot *ms_slot;
	unsigned long *l_ptr;
	unsigned char TrackNummer;
	unsigned char Start;
	unsigned char * w_buffer = (unsigned char *) W_BUFFER_ADR;	
	unsigned int *NameBuffer 	= (unsigned int *) 	NAMEBUFFER_ADR;
	unsigned char *DirBuffer	= (unsigned char *) DIRBUFFER_ADR;
	unsigned char * entry;

	entry = (unsigned char*) SECTORBUFFER_ADR;
	read_lba(d_sector, 1, entry);	//read directory sector

	dummy =-1;
	while(++dummy < 16)							//max 16 entries
	{
		switch( *entry)
		{	case 0:		return 1;				//last entry ?
			case 0x2E:							//Dir: . or .. we don't need
			case 0xE5:							//File deleted
				*w_buffer = 0x00;
				continue;
		}

		if( (entry[11] & ATTR_SLOT) == ATTR_SLOT)	//Win95 DirSlot ?
		{
//			ms_slot =(struct msdos_dir_slot *)&entry[dummy];
			strcpy((unsigned char *)W1_BUFFER_ADR, w_buffer);			//save old buffer				
			ws_strncpy( w_buffer,  (unsigned int*)&ms_slot->name0_4[0], 5);		//copy ldn name 0-4
			ws_strncpy( &w_buffer[5],  (unsigned int*)&ms_slot->name5_10[0], 6);//copy ldn name 5-10
			ws_strncpy( &w_buffer[11], (unsigned int*)&ms_slot->name11_12[0], 2);//copy ldn name 11-12
			strcpy(&w_buffer[13],(unsigned char *)W1_BUFFER_ADR);
			continue;
		}
		
		if( (entry[dummy].attr & (ATTR_HIDDEN | ATTR_VOLUME)) != 0)	
		{
			*w_buffer = 0x00;
			continue;
		}

		if( *w_buffer == 0x00)								//long filename check
			strncpy(w_buffer, entry[dummy].name,11);			//Dos-Name kopieren

		l_ptr = (unsigned long *) &DirBuffer[SaveOffset];
		l_ptr[0] = ((unsigned long) entry[dummy].starthi <<16) | entry[dummy].start; //save Cluster
		l_ptr[1] = entry[dummy].size;			//save size

		if (strcmp(w_buffer,"playlist.m3u") == 0)			//Vergleich mit dem Namen "playlist.m3u"
		{	PlaylistStartCluster = l_ptr[0];				//Namen sind gleich, also Start speichern
			PlaylistLaenge =  l_ptr[1];						//Lnge speichern
			continue;										//von vorn weiter machen
		}
		
/*		if ((w_buffer[0] > 0x39) || (w_buffer[0] < 0x30) ||
			(w_buffer[1] > 0x39) || (w_buffer[1] < 0x30)) 
		{	
			sendstring("\n\r? ->");		
			sendstring(w_buffer);		
			return 2;
		}
*/
		TrackNummer = (w_buffer[0] - 0x30) *10 + w_buffer[1] - 0x31;		

		NameBuffer[TrackNummer] = SaveOffset;					//im Namebuffer die Anfangsadresse 
																//von diesem Eintrag festhalten
		SaveOffset += (2* sizeof(unsigned long));																				
		
		if((entry[dummy].attr & ATTR_DIR) == ATTR_DIR)			//Wenn Verzeichnis:
		{  	
			Start = 2;											//die Zahl im Namen ausblenden
			w_buffer[2] = DIR_MARK;								//Kennung anbringen
		}
		else 
		{	w_buffer[strlen(w_buffer)-4] = 0;					// die Extension .mp3 entfernen	
			Start = 0;
		}
		
		strcpy( &DirBuffer[SaveOffset], &w_buffer[Start]);		//save filename
		SaveOffset += strlen(w_buffer) +1;						//+1 for 0
		*w_buffer = 0x00;										//abschlieende 0 einbringen
		DirCount++;												//ein Eintrag mehr vorhanden
	}
	return 0;		//noch kein Ende von dieser Funktion in Sicht, nchster Sektor
}

/***********************************************************************************
- Startet das Einlesen einen Verzeichnisses
- liest alle Sektoren eines Verzeichnisseintrages
***********************************************************************************/
void dir( unsigned long cluster)
{
	unsigned int sec_to_read;
	unsigned long first_sector;

	DirCount = 0;
	SaveOffset = 0;
	PlaylistLaenge = 0;
	*(unsigned char *)W_BUFFER_ADR =0;
	memset((unsigned char *)NAMEBUFFER_ADR, 0xff, NAMEBUF_S);

	if (cluster == 0)
	{
		first_sector = DIR_START;		//read ROOT dir
		sec_to_read = Fat32Length;
	}
	else
	{
		first_sector = cluster_to_sector( cluster);		//read subdir
		sec_to_read = ClusterSize;
	}

	while (ZeitFenster == 0);	//Warten bis PlayBuffer	upgedated
	while(1)
	{	unsigned int sector;
		for(sector = 0; sector < sec_to_read; sector++)
		{
			if  ( read_fat_sector( first_sector + sector))	return ;
		}		
		cluster = get_next_cluster(cluster);			 
		first_sector = cluster_to_sector(cluster);
	}
}

/***********************************************************************************
- berechnet die erste Sectornummer aus der Clusternummer
***********************************************************************************/
unsigned long cluster_to_sector( unsigned long cluster)
{
//	return  DataStart + (cluster-2) * ClusterSize;
	return  DataStart + ((cluster-2) <<4);
}

/***********************************************************************************
- ein Sektor der Fat wird gelesen und in den FatBuffer geschrieben
***********************************************************************************/
#define FAT32_ENTRYS_SECTOR	(SECTORSIZE / 4)
void get_fat( unsigned long cluster)
{
	read_lba( FAT_START + (cluster>>7), (FAT_BUF_S/SECTORSIZE), (unsigned char *)FATBUFFER_ADR);
}

/***********************************************************************************
- es wird aus der Clusterkette der nchste Cluster gesucht ber die FAT
- wenn der Cluster nicht im FatBuffer steht, wird der ensprechende Teil aus der Fat geladen
***********************************************************************************/

#define ENTRY_CACHE_32	(FAT_BUF_S / 4)

unsigned long get_next_cluster( unsigned long prev_cluster)
{
unsigned long *FatBuffer	= (unsigned long *)	FATBUFFER_ADR;

	if	((prev_cluster >= (ENTRY_CACHE_32 + (first_cluster_in_cache)))||
		(prev_cluster <  first_cluster_in_cache ))
	{
		get_fat( prev_cluster);	//wenn angeforderter Cluster nicht im Cache -> hole FAT
		first_cluster_in_cache = prev_cluster & 0xffffff80;
	}
	return (FatBuffer[prev_cluster - first_cluster_in_cache]);
}

/***********************************************************************************
- es wird ein Cluster eingelesen und in den Playbuffer geschrieben
***********************************************************************************/
void update_buf(unsigned char *buffer )
{
	read_lba(cluster_to_sector(last_cluster),ClusterSize, buffer);
	last_cluster = get_next_cluster(last_cluster);

/*	if (ClusterSize == 8)
	{	read_lba(cluster_to_sector(last_cluster),ClusterSize, &buffer[0x1000]);
		last_cluster = get_next_cluster(last_cluster);
	}
*/	
	ZeitFenster = 1;	//ein kurzes Zeitfenster aufmachen, wird im TimerInt wieder gelscht		
}

